from internetnl_domain_analyse.utils import prepare_stat_data_for_write

__author__ = "Eelco van Vliet"
__copyright__ = "Eelco van Vliet"
__license__ = "MIT"


def test_prepare_data():

    assert prepare_stat_data_for_write is not None
